"""Query SAP (service access point) information from Alcatel 5620 SAM."""

import optparse
import pprint
import re
import sys
import traceback

import sam.sam


class Querier(object):
    """Querier of service information from SAM."""

    option_attrib_map = {
        'port': 'portName',
        'device': 'nodeName',
        'node': 'nodeName',
        'id': 'serviceId',
        'name': 'serviceName',
        'type': 'serviceType',
        'description': 'description',
        'subname': 'subscriberName',
        'subid': 'subscriberId',
        'innertag': 'innerEncapValue',
        'outertag': 'outerEncapvalue',
        }

    attrib_map = {}
    for k, v in option_attrib_map.iteritems():
        attrib_map[v] = k

    def __init__(self, host, username, password=None, digest=None, port=8080):
        if password is None and digest is None:
            raise ValueError(
                'Either a cleartext or digest password must be supplied.')
        elif password is not None:
            args = dict(password=password)
        elif digest is not None:
            args = dict(password_digest=digest)
        self.sam = sam.sam.SAM(
            host, port=port, username=username, **args)

    def handle_options(self, options):
        r = {}
        for key in options:
            keyname = self.option_attrib_map.get(key)
            if keyname is not None and options.get(key):
                r[keyname] = options.get(key)
        return r

    def _query(self, **args):
        if args.get('description'):
            # Copy the value and remove it from the dict
            desc = args['description'][:]
            del args['description']
        else:
            desc = None
        saps = self.sam.saps_extended(**args)
        # Now allow regexp searching by description
        result = []
        if desc:
            desc_re = re.compile(desc)
        for sap in saps:
            if desc is None or desc_re.match(sap.description):
                result.append(sap)
        return result

    def display_services(self, **args):
        args = self.handle_options(args)
        saps = self._query(**args)

        for sap in saps:
            for k, v in sap.iteritems():
                print '%s: %s' % (k, v)
            print '-' * 60


def get_cmdline_args():
    parser = optparse.OptionParser()
    # Connection params (move to rc-file)
    parser.add_option(
        '--host', dest='host', help='SAM-O hostname')
    parser.add_option(
        '-u', '--username', dest='username', help='SAM-O username')
    parser.add_option(
        '--port', dest='hostport', help='SAM-O port', default=8080)
    parser.add_option(
        '--password', dest='password', help='SAM-O password')
    parser.add_option(
        '--digest', dest='digest', help='SAM-O password digest')

    # Query parameters
    parser.add_option(
        '--int', '--interface', dest='port',
        help='Device interface/port the SAP is bound to')
    parser.add_option(
        '--node', dest='node',
        help='Device name the SAP lives on')
    parser.add_option(
        '--device', dest='device',
        help='Device name the SAP lives on')
    parser.add_option(
        '--id', dest='id',
        help='Service ID')
    parser.add_option(
        '--name', dest='name',
        help='Service name')
    parser.add_option(
        '-t', '--type', dest='type',
        help='Service type (use "vll", "epipe", "vpls", "vprn")')
    parser.add_option(
        '-d', '--description', dest='description',
        help='Service description (may use a regular expression)')
    parser.add_option(
        '--subid', dest='subid',
        help='Subscriber ID')
    parser.add_option(
        '--subname', dest='subname',
        help='Subscriber Name')
    parser.add_option(
        '--innertag', dest='innertag',
        help='Inner VLAN tag on a QinQ service')
    parser.add_option(
        '--outertag', dest='outertag',
        help='Outer VLAN tag on a QinQ service')

    options, args = parser.parse_args()
    return (options, args)


def main():
    o, _  = get_cmdline_args()
    try:
        q = Querier(o.host, o.username, o.password, o.digest, port=o.hostport)
    except Exception, e:
        sys.stderr.write('Error: %s' % str(e))
    else:
        try:
            q.display_services(
                port=o.port, node=o.node, device=o.device, id=o.id, name=o.name,
                type=o.type, description=o.description, innertag=o.innertag,
                outertag=o.outertag, subid=o.subid, subname=o.subname)
        except sam.sam.XMLException, e:
            sys.stderr.write('Error: %s' % str(e))
        except Exception, e:
            sys.stderr.write('Error: %s' % str(e))
            sys.stderr.write(traceback.format_exc())
            
        

if __name__ == '__main__':
    main()
