"""Alcatel 5620 SAM Python interface.

This module provides a simple interface for querying Alcatel's 5620
Service Aware Manager. It takes care of the rather grotesque SOAP
interface, allowing to make queries and giving you answers in the
format you know and love - Python dictionaries.

Originally written by Rod Telford.  ElementTree XML handling,
templating using standard library, generic queries and other changes
by Andrew Fort.
"""

__author__ = 'Andrew Fort <a@choqolat.org>'
__author__ = 'Rod Telford <rod@dailysnore.com>'

try:
    import hashlib
except:
    import md5 as hashlib

import httplib
import string
try:
    from xml.etree import cElementTree  # 2.5+
except:
    import cElementTree  # 2.4

import random
import xml.sax.saxutils

import const
import xmldict


class Error(Exception):
    pass


class FindException(Error):
    """An error occured during a SAM query."""


class XMLException(Error):
    """There was an error in some XML sent to or from SAM."""


class SAM(object):
    """Alcatel 5620 Service Aware Manager Interface for Python.

    This module interfaces with the 5620 SAM-O SOAP interface. Since SOAP
    is meant to be as painful as possible and make your eyes bleed, this
    module is intended to restore your sanity.

    Attributes:
      timeout: A float, the SAM-O SOAP request timeout (in seconds).
        When this value is changed, the timeout will be reflected on
        the next request.
    """

    def __init__(self, host, port=8080, username='', password=None,
                 password_digest=None, timeout=const.DEFAULT_TIMEOUT):
        """Initialiser.

        Use either password or password_digest, but not both.

        Args:
          host: A string, the hostname or IP address of the SAM-O server.
          port: An int, the port number of the SAM-O server.
          username: A string, the username to use in the SOAP request.
          password: A string, the unencrypted password (will be hashed).
          password_digest: A string, a pre-hashed password already in
            digest form.
          timeout: A float, the SAM-O request timeout (in seconds).
        """
        self.timeout = timeout
        self.username = username
        self._host = host
        self._port = port
        if password_digest is not None:
            self._password_digest = password_digest
        elif password is not None:
            self._password_digest = hashlib.md5(password).hexdigest()
        else:
            raise ValueError(
                'Either password or password_digest argument must be supplied')

    def ports(self, site_id=None, port_name=None):
        """Requests Physical Ports, optionally filtered.

        Args:
          site_id: A string, an IP address to filter the query by.
          port_name: A string, a port name to filter the query by.

        Returns:
          A list of dictionaries, physical port results.

        Raises:
          FindException if there was some error in the query.
          XMLException if SAM-O didn't like the XML (unlikely).
        """
        d = {}
        if site_id is not None:
            d['siteId'] = site_id
        if port_name is not None:
            d['portName'] = 'Port %s' % port_name
        result = self.generic_query(const.PORT_CLASS, query=d,
                                    attribs=const.PORT_GENERIC_RESULT_FILTER)
        r = []
        if not result:
            return r
        for res_list in result.itervalues():
            if isinstance(res_list, dict):
                r.append(res_list)
            else:
                r.extend(res_list)
        return r

    def saps(self, node_name=None, port_name=None):
        """Requests Service Access Point information, optionally filtered.

        Args:
          node_name: A string, the node name to filter by (e.g., 'ar1.syd')
          port_name: A string, the port name to filter by (e.g., '1/2/1')

        Returns:
          A list of dictionaries, service access results.

        Raises:
          FindException if there was some error in the query.
          XMLException if SAM-O didn't like the XML (unlikely).
        """
        d = {}
        if node_name is not None:
            d['nodeName'] = node_name
        if port_name is not None:
            d['portName'] = 'Port %s' % port_name
        result = self.generic_query(const.SAP_CLASS, query=d,
                                    attribs=const.SAP_GENERIC_RESULT_FILTER)
        r = []
        if not result:
            return r
        for res_list in result.itervalues():
            if isinstance(res_list, dict):
                r.append(res_list)
            else:
                r.extend(res_list)
        return r

    def saps_extended(self, **query):
        """Extended Service Access Point request.

        Args:
          Keyword arguments as understood by SAM, e.g., "siteId='172.16.0.1'".
          One may also use a dictionary prefixed by **.

        Returns:
          A list of SAP records (dictionaries).
        """
        if query:
            d = query
        else:
            d = None

        result = self.generic_query(const.SAP_CLASS, query=d,
                                    attribs=const.SAP_GENERIC_RESULT_FILTER)
        r = []
        if not result:
            return r
        for res_list in result.itervalues():
            if isinstance(res_list, dict):
                r.append(res_list)
            else:
                r.extend(res_list)
        return r       
    
    def generic_query(self, obj, query=None, attribs=None, no_children=False):
        """Requests for an arbitrary object and device query.

        Args:
          obj: A string, the full class of the object to query SAM-O for.
          query: A dictionary, key/value pairs to use as a query filter.
          attribs: A list of strings, attribute names to include in the
            response dictionary.
          no_children: A boolean; if True, do not return any children groups
            in the response (can reduce the response size dramatically).

        Returns:
          A dictionary, containing recursively nested dictionaries or
          lists of dictionaries.

        Raises:
          FindException if there was some error in the query.
          XMLException if SAM-O didn't like the XML (unlikely).
        """
        if query is None:
            query = {}
        q = self._query(obj, query, attribs, no_children=no_children)
        result = self._request(q)
        d = self._dictify_sam_response(result)
        if d.get('Body').get('findException'):
            exc = d.get('Body').get('findException').get('description')
            raise FindException(exc)
        elif d.get('Body').get('XMLException'):
            exc_info = ['%s=%s' % (k, v) for
                        (k, v) in d.get('Body').get('XMLException').iteritems()]
            exc = 'XML Error. %s. XML request was:\n %s' % (', '.join(exc_info),
                                                            q)
            raise XMLException(exc)
        else:
            if d.get('Body').get('findResponse') is None:
                raise FindException('No findResponse object in response')
            elif d.get('Body').get('findResponse').get('result', -1) == -1:
                raise FindException(
                    'No response objects found in response from SAM')
            else:
                return d.get('Body').get('findResponse').get('result')

    def _query(self, full_class, request_filter=None, response_attribs=None,
               no_children=False):
        """Builds a generic SAM request, with filtering."""
        r = const.SOAP_START + const.FIND_START + (
            string.Template(const.FULL_CLASS).substitute(cls=full_class))
        if request_filter:
            f = string.Template(const.FILTER)
            fe = []
            if no_children:
                childclause = const.NO_CHILDREN
            else:
                childclause = const.CHILDREN
            for k, v in request_filter.items():
                _f = string.Template(const.FILTER_ELEMENT)
                k = xml.sax.saxutils.escape(k)
                v = xml.sax.saxutils.escape(v)
                fe.append(_f.substitute(name=k, value=v))
            r += f.substitute(cls=full_class, filter=''.join(fe))

        if response_attribs or no_children:
            f = string.Template(const.RESULT_FILTER)
            fe = []
            if no_children:
                childclause = const.NO_CHILDREN
            else:
                childclause = const.CHILDREN
            if response_attribs:
                for attr in response_attribs:
                    _f = string.Template(const.RESULT_ATTRIB)
                    fe.append(_f.substitute(attrib=attr))
            r += f.substitute(filter=''.join(fe), children=childclause)
        r += const.FIND_END + const.SOAP_END
        return r

    def _dictify_sam_response(self, xml_data,
                              key_regexp=None, value_regexp=None):
        """Converts the XML response into a python dictionary."""
        tree = xmldict.convert_xml_to_dict(xml_data)
        return tree.get('Envelope') or {}

    def _request(self, soap_req, timeout=None, uri='/xmlapi/invoke'):
        """Sends the SOAP request out on the wire."""
        req = self._request_string(soap_req)
        timeout = timeout or self.timeout
        ws = httplib.HTTPConnection(self._host, self._port, timeout=timeout)
        ws.putrequest('POST', uri)
        ws.putheader('User-Agent', 'pySAM')
        ws.putheader('Content-type', 'text/xml: charset="UTF-8"')
        ws.putheader('Content-length', '%d' % len(req))
        ws.putheader('SOAPAction', '""')
        ws.endheaders()
        ws.send(req)
        resp = ws.getresponse()
        if resp.status == 200:
            data = resp.read()
            ws.close()
            return data
        else:
            ws.close()
            raise ValueError((resp.status, resp.reason))

    def _request_string(self, xml_body):
        """Wraps the XML request body in the SOAP envelope."""
        t = string.Template(const.SAMO_REQUEST_ENVELOPE)
        return t.substitute(username=self.username,
                            password=self._password_digest,
                            request=xml_body)
